// Using LittleFS/SPIFFS. Using a file(ascii+nonascii) to load/store variables, data, parameters. Not thread safe. by ken woo. v.0.4.
//
// * version 0.0. 20210319. initial version.
//
// * version 0.1. 20210320.
// change some names and reorder others for consistency.
// add delete record function.
// add renew function. because the deleted record ones still exist in file. renew is clone a clean one and remove old.
// add list records names and ids.
//
// * version 0.2. 20210321.
// fixed a bug in my_find.
// add demo main function.
//
// * version 0.3. 20210323.
// refine some existing functions for more usable.
// fixed WriteBytes since yet the buffer/data length both need specified.
// function setup_filesystem and its counter are more suitable to put in open/close functions.
// add remove file(auto close file if opened), and format(viable when no file opened) functions.
// add functions FindIdNextEx and FindMaxIdEx and Cp and ReplaceWordsEx and ReplaceBytesEx and ReadData.
//
// * version 0.4. 20210402.
// fixed a minor bug in Renew().
// arranged into an example and remove the .cpp and only keep a .h.
//
// # dev NOTE: since File handler seems not trivial, another var fmode is used, consistency unsured.
//
// # NOTE: DeleteRecords/Renew/ReplaceWordsEx/ReplaceBytesEx methods are only viable when file is at the Closed() state;
// since fopen mode is different while applied.
//
// % to do: but unable to do, rtti for typeinfo so as to recognize data type just via db file.


#include <ESP8266WiFi.h>
#include<LittleFS.h>


#ifndef _c_VARS_2_FILE_H_
#define _c_VARS_2_FILE_H_


#define FILESYSTEM LittleFS
//#define FILESYSTEM SPIFFS // its format() takes too long.


typedef class cDataBase{

    #define PREAMBLE        (0x55A55AAA)
    #define DEPREAMBLE      (0x55A55AAB)
    // a complete record unit in a file:
    // [4-byte preamble]+[4-byte name length]+[4-byte name checksum]+[4-byte record length]+[n-byte name]+[m-byte record contents]=16+m+n.
    // payload is 16kbytes/1000records.

    typedef struct __attribute__((__packed__)) sRecord{
        unsigned nlen; // name length.
        unsigned ncs; // name checksum.
        unsigned rlen; // record length, name+id+data.
    } sRecord;

    static void setup_filesystem(){
        if (!filesystem_opened){ // load FS whenever file opened, unload whenever no file opened. counter used.
            FILESYSTEM.begin();
        }
        filesystem_opened++;

        static int x=1;
        if (x){
            x=0;
            {
                Dir dir = FILESYSTEM.openDir("/");
                if (!dir.next()){
                    FILESYSTEM.format();
                    printf("\r\n\r\nfilesystem formatted.\r\n");
                }
            }
            FSInfo fs_info;
            if (FILESYSTEM.info(fs_info)){
                printf(\
                    "--------\r\nThe FileSystem:\r\nentire space %d bytes\r\n%d bytes used\r\n"\
                    "free space %d bytes\r\nblock size %d bytes\r\npage size %d blocks\r\n--------\r\n",\
                    fs_info.totalBytes, fs_info.usedBytes, fs_info.totalBytes-fs_info.usedBytes, fs_info.blockSize, fs_info.pageSize);
            }
            Dir dir = FILESYSTEM.openDir("/");
            printf("list root directory \"/\":\r\n");
            while (dir.next())
                printf("%s    %s\r\n", dir.fileName().c_str(), dir.isDirectory()? "<DIR>": (String(dir.fileSize())+" bytes").c_str());
        }
    };

    unsigned my_find(const void *str, unsigned sz){ // only for either char or unsigned char. 0/not found. others/the found position.
        unsigned i=0, j=0, w=sz*2, sum=0, csum=0, start_p=0, end_p=sz;
        unsigned char u[w+sz];
        int v=0;

        if (fh.readBytes((char*)u, sz)!=sz) return 0;
        for (; i<sz; i++){
            sum+=((const unsigned char*)str)[i];
            csum+=u[i];
        }

        while (1){
            if (sum==csum){
                for (i=0, j=0; i<sz && (((const unsigned char*)str)[i]==u[start_p+i-j]); i++) if (start_p+i-j==sz+w-1) j=start_p+i+1;
                if (i==sz){
                    fh.seek(-v, SeekCur);
                    return fh.position();
                }
            }
            if (!v){
                if (end_p==sz) v=fh.readBytes((char*)(u+end_p), w);
                else if (end_p==w+sz) v=fh.readBytes((char*)u, w);
                else {v=fh.readBytes((char*)(u+w), sz); v+=fh.readBytes((char*)u, sz);}
                if (!v) return v;
            }
            if (end_p==w+sz) end_p=0;
            csum+=u[end_p++];
            csum-=u[start_p++];
            if (start_p==w+sz) start_p=0;
            v--;
        }
    };

    static unsigned filesystem_opened;
    const unsigned preamble=PREAMBLE;
    const unsigned depreamble=DEPREAMBLE;

    const char *db_name;
    const char *fmode;
    File fh;

    public:

    cDataBase(): db_name(0), fmode(0){;};
    cDataBase(const char *dbn): db_name(dbn), fmode(0){;};
    void operator=(const char *dbn){
        this->Close();
        db_name=dbn;
    };
    ~cDataBase(){ // may not run immediately, so the counter should not put here.
        this->Close();
    };

    bool Open(const char *m="a+"){ // it now becomes important; must a file opened to load FS.
        if (!db_name || fmode) return false;
        setup_filesystem();
        fh=FILESYSTEM.open(db_name, m);
        if (fh){fmode=m; return true;}
        else if (!--filesystem_opened) FILESYSTEM.end();
        return false;
    };

    void Close(){
        if (fmode){
            fh.close();
            fmode=0;
            if (!--filesystem_opened) FILESYSTEM.end();
        }
    };

    void DumpAll(int f=0){ // 0: ascii; 1: hex.
        if (!db_name) return;
        unsigned fpos=0, opened=1;
        if (fmode){
            if (((*fmode=='w') || (*fmode=='a')) && !fmode[1]) return;
            fpos=fh.position();
            fh.seek(0, SeekSet);
        }
        else {
            if (!this->Open("r")) return;
            opened=0;
        }

        char u[8];
        int v;
        unsigned counter=0, checksum=0;

        while ((v=fh.readBytes(u, 8))==8){
            counter+=8;
            printf("[");
            for (int i=0; i<8; i++){
                checksum+=u[i];
                if (f) printf("%02X ", u[i]);
                else {
                    if (u[i]<' ' || u[i]>'~') u[i]='.';
                    printf("%2c ", u[i]);
                }
            }
            printf("%s\r\n", "]");
        }
        counter+=v;
        printf("[");
        for (int i=0; i<v; i++){
            checksum+=u[i];
            if (f) printf("%02X ", u[i]);
            else {
                if (u[i]<' ' || u[i]>'~') u[i]='.';
                printf("%2c ", u[i]);
            }
        }
        printf("%s\r\n", "]");
        printf("[%d bytes, checksum=%d]\r\n", counter, checksum);

        if (!opened) this->Close();
        else fh.seek(fpos, SeekSet);
    };

    template<typename T>
    void WriteWords(const char *name, int id, T data){
        // for elementary types, id whether unique or not maintained by user.
        // name is for the db to search purpose; whether unique or not maintained by user.

        sRecord u;

        for (u.nlen=0, u.ncs=0; name[u.nlen]; u.ncs+=name[u.nlen], u.nlen++);
        u.rlen=u.nlen+sizeof(id)+sizeof(data);

        fh.write((const unsigned char*)&preamble, sizeof(preamble));
        fh.write((const unsigned char*)&u, sizeof(u));
        fh.write((const unsigned char*)name, u.nlen);
        fh.write((unsigned char*)&id, sizeof(id));
        fh.write((unsigned char*)&data, sizeof(data));
    };

    template<typename T>
    void WriteBytes(const char *name, int id, const T *buffer, unsigned buff_len, unsigned total_len){
        sRecord u;

        for (u.nlen=0, u.ncs=0; name[u.nlen]; u.ncs+=name[u.nlen], u.nlen++);
        u.rlen=u.nlen+sizeof(id)+sizeof(T)*total_len; // note, not sizeof(buffer).

        fh.write((const unsigned char*)&preamble, sizeof(preamble));
        fh.write((const unsigned char*)&u, sizeof(u));
        fh.write((const unsigned char*)name, u.nlen);
        fh.write((unsigned char*)&id, sizeof(id));
        fh.write((const unsigned char*)buffer, sizeof(T)*buff_len);
    };

    template<typename T>
    void WriteBytesCont(const T *buffer, unsigned length){ // only runs along with WriteBytes
        fh.write((const unsigned char*)buffer, sizeof(T)*length);
    };

    void FindStart(){ // a brand new start to find; from file ptr position 0.
        fh.flush();
        fh.seek(0, SeekSet);
    };

    unsigned FindNext(const char *name, int id=-1, unsigned proc=0){
        // find start from the current file position. file pos is moving up with FindNext.
        // return the fpos points to the begin of data(id+data contents) when id==-1. (and the fpos is where the current fptr pointing to).
        // return the fpos points to the begin of data contents when id!=-1. (and the fpos is where the current fptr pointing to).
        //
        // later, the var proc added:
        // 0/default, 1/delete when name(or and id) matched.
        // special!!! 2/for list name and id. because lacks target can stored the info, the returned var is the name length.
        // id though, is 4-byte but may not proper when directly specifies 4 but do.
        // 3/for renew. again, move fpos to preamble head and return the entire record length including preamble.
        // 4/for search same id without specifying name; result is name pos and its length.
        // 5/for search same name and id, positioning at data head and return data length.

        unsigned sum=0, i=0, len=0, pream=0;
        sRecord u;
        for (; name[len]; len++) sum+=name[len];

        while (pream=my_find(&preamble, sizeof(preamble))){
            if (fh.readBytes((char*)&u, sizeof(u))==sizeof(u)){
                if ((len==u.nlen) && (sum==u.ncs)){
                    char n[u.nlen];
                    if (fh.readBytes(n, u.nlen)==u.nlen){
                        for (i=0; i<u.nlen && (name[i]==n[i]); i++);
                        if (i==u.nlen){
                            if (id!=-1){
                                unsigned rid;
                                if (fh.readBytes((char*)&rid, sizeof(rid))==sizeof(rid)){
                                    if (id!=rid) fh.seek(u.rlen-u.nlen-sizeof(rid), SeekCur);
                                    else {
                                        if (proc==1){ // delete. move pos to the preamble head.
                                            fh.seek(pream-sizeof(preamble), SeekSet);
                                            fh.write((const unsigned char*)&depreamble, sizeof(depreamble));
                                            return fh.position();
                                        }
                                        else if (proc==5){
                                            return u.rlen-u.nlen-sizeof(id);
                                        }
                                        else return fh.position(); // found data
                                    }
                                }
                                else {/*fh.seek(0, SeekEnd); */return 0;}
                            }
                            else {
                                if (proc==1){ // delete. move pos to the preamble head.
                                    fh.seek(pream-sizeof(preamble), SeekSet);
                                    fh.write((const unsigned char*)&depreamble, sizeof(depreamble));
                                    return fh.position();
                                }
                                else return fh.position(); // found id+data
                            }
                        }
                        else fh.seek(u.rlen-u.nlen, SeekCur);
                    }
                    else {/*fh.seek(0, SeekEnd); */return 0;}
                }
                else if ((proc==2) && (id==-1)){ // list name and id
                    return u.nlen;
                }
                else if ((proc==3) && (id==-1)){ // renew
                    fh.seek(pream-sizeof(preamble), SeekSet);
                    return sizeof(preamble)+sizeof(sRecord)+u.rlen;
                }
                else if ((proc==4) && (id!=-1)){ // search for id only
                    unsigned rid;
                    fh.seek(u.nlen, SeekCur);
                    if (fh.readBytes((char*)&rid, sizeof(rid))==sizeof(rid)){
                        if (rid==id){
                            fh.seek(-(u.nlen+sizeof(rid)), SeekCur);
                            return u.nlen; // found name+id+data, return name length.
                        }
                        else fh.seek(u.rlen-u.nlen-sizeof(rid), SeekCur);
                    }
                    else {/*fh.seek(0, SeekEnd); */return 0;}
                }
                else fh.seek(u.rlen, SeekCur);
            }
            else {/*fh.seek(0, SeekEnd); */return 0;}
        }
        /*fh.seek(0, SeekEnd);*/
        return 0;
    };

    template<typename T>
    bool ReadWords(T &data, int &id, const char *name){ // read both the data and id according to name.
        FindStart();
        if (FindNext(name)){
            if (fh.readBytes((char*)&id, sizeof(id))==sizeof(id)){
                if (fh.readBytes((char*)&data, sizeof(data))==sizeof(data)) return true;
                return false;
            }
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadWords(T &data, const char *name, int id=-1){ // read the data only according to name or more with the id when id!=-1.
        FindStart();
        if (FindNext(name, id)){
            if (id==-1) fh.seek(sizeof(id), SeekCur);
            if (fh.readBytes((char*)&data, sizeof(data))==sizeof(data)) return true;
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadWordsNext(T &data, int &id, const char *name){ // read both the data and id according to name. Next means start from current pos.
        if (FindNext(name)){
            if (fh.readBytes((char*)&id, sizeof(id))==sizeof(id)){
                if (fh.readBytes((char*)&data, sizeof(data))==sizeof(data)) return true;
                return false;
            }
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadWordsNext(T &data, const char *name, int id=-1){ // read the data only according to name or more with the id when id!=-1.
        if (FindNext(name, id)){
            if (id==-1) fh.seek(sizeof(id), SeekCur);
            if (fh.readBytes((char*)&data, sizeof(data))==sizeof(data)) return true;
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadBytes(T *buffer, unsigned length, int &id, const char *name){ // read both the data and id according to name.
        FindStart();
        if (FindNext(name)){
            if (fh.readBytes((char*)&id, sizeof(id))==sizeof(id)){
                if (fh.readBytes((char*)buffer, sizeof(T)*length)==sizeof(T)*length) return true; // note, not sizeof(buffer).
                return false;
            }
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadBytes(T *buffer, unsigned length, const char *name, int id=-1){
    // read the data only according to name or more with the id when id!=-1.

        FindStart();
        if (FindNext(name, id)){
            if (id==-1) fh.seek(sizeof(id), SeekCur);
            if (fh.readBytes((char*)buffer, sizeof(T)*length)==sizeof(T)*length) return true; // note, not sizeof(buffer).
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadBytesNext(T *buffer, unsigned length, int &id, const char *name){ // read both the data and id according to name.
        if (FindNext(name)){
            if (fh.readBytes((char*)&id, sizeof(id))==sizeof(id)){
                if (fh.readBytes((char*)buffer, sizeof(T)*length)==sizeof(T)*length) return true; // note, not sizeof(buffer).
                return false;
            }
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadBytesNext(T *buffer, unsigned length, const char *name, int id=-1){
    // read the data only according to name or more with the id when id!=-1.

        if (FindNext(name, id)){
            if (id==-1) fh.seek(sizeof(id), SeekCur);
            if (fh.readBytes((char*)buffer, sizeof(T)*length)==sizeof(T)*length) return true; // note, not sizeof(buffer).
            return false;
        }
        return false;
    };

    template<typename T>
    void ReadBytesCont(T *buffer, unsigned length){ // only runs along with ReadBytes/ReadBytesNext
        fh.readBytes((char*)buffer, sizeof(T)*length);
    };

    unsigned ReadData(unsigned char *buffer, unsigned buff_len, const char *name, int id){ // used when data length is unknown.
    // read the data according to both the name and id, only the 1st occurrence is success.
    // the buffer will be filled if data over and will return the total data length for succeeding processing.

        unsigned sz, i;
        if (id<0) return 0;
        FindStart();
        if (i=sz=FindNext(name, id, 5)){
            if (buff_len<i) i=buff_len;
            if (fh.readBytes((char*)buffer, i)!=i) return 0;
        }
        return sz;
    };

    unsigned ReadDataCont(unsigned char *buffer, unsigned buff_len){ // only used along with ReadData. number of bytes read returned.
    // how many data for read must controlled by user. that is, according to the returned value of ReadData previously called.
        return fh.readBytes((char*)buffer, buff_len);
    };

    bool LsRecords(){
        if (!db_name) return false;
        unsigned fpos=0, opened=1, len=0;
        if (fmode){
            if (((*fmode=='w') || (*fmode=='a')) && !fmode[1]) return false;
            fpos=fh.position();
            fh.seek(0, SeekSet);
        }
        else {
            if (!this->Open("r")) return false;
            opened=0;
        }
        FindStart();
        printf("%s record contents:\r\n--------\r\n", db_name);
        while (len=FindNext("", -1, 2)){
            {
                char r[len+1];
                unsigned i;
                r[len]=0;
                fh.readBytes(r, len);
                fh.readBytes((char*)&i, 4);
                printf("[%s    id=%u]\r\n", r, i);
            }
        }
        printf("--------\r\n");
        if (!opened) this->Close();
        else fh.seek(fpos, SeekSet);
        return true;
    };

    bool DeleteRecords(const char *name, int id=-1){ // delete the data according to the name(or and id), all of which are deleted.
        if (!db_name || fmode || !this->Open("r+")) return false;
        FindStart();
        bool r=FindNext(name, id, 1);
        while (r && FindNext(name, id, 1));
        this->Close();
        return r;
    };

    bool Renew(){
        if (!db_name || fmode || !this->Open("r")) return false; // make sure at least a file opened.
        int i;
        for (i=0; i<20 && FILESYSTEM.exists(String("/tmp")+i); i++);
        if (i==20){ // no more try.
            this->Close();
            return false;
        }
        String newfile=String("/tmp")+i;
        File tmp=FILESYSTEM.open(newfile, "w");
        if (!tmp){
            this->Close();
            return false;
        }
        FindStart();
        unsigned r=FindNext("", -1, 3);
        if (!r){
            tmp.close();
            FILESYSTEM.remove(newfile.c_str());
            this->Close(); // not reasonable but must after tmp closed and removed.
            return false;
        }
        do { //// no care about full.
            {
                char s[r];
                fh.readBytes(s, r);
                tmp.write((unsigned char*)s, r);
            }
        }
        while (r=FindNext("", -1, 3));
        tmp.close();
        this->Close(); // not reasonable but must after tmp closed.

        if (!filesystem_opened) FILESYSTEM.begin(); // not reasonable but must.
        FILESYSTEM.remove(db_name);
        FILESYSTEM.rename(newfile.c_str(), db_name);
        if (!filesystem_opened) FILESYSTEM.end(); // not reasonable but must.
        return true;
    };

    bool Remove(){
        if (!db_name) return false;
        this->Close();
        if (!filesystem_opened) FILESYSTEM.begin(); // not reasonable but must.
        FILESYSTEM.remove(db_name);
        if (!filesystem_opened) FILESYSTEM.end(); // not reasonable but must.
        db_name=0;
        return true;
    };

    static bool Format(){ // only works if no file opened.
        if (!filesystem_opened) FILESYSTEM.begin();
        else return false;
        FILESYSTEM.format();
        FILESYSTEM.end();
        return true;
    };

    template<typename T>
    bool ReplaceWordsEx(const char *name, int id, T data){ // replace or add the record.
        // suppose there are not only a matched records, only 1st could success. in this case, del all then add instead.
        if (!db_name || fmode || !this->Open("r+") || (id==-1)) return false;
        FindStart();
        bool r=FindNext(name, id, 1);
        while (r && FindNext(name, id, 1));
        fh.seek(0, SeekEnd);
        WriteWords(name, id, data);
        this->Close(); // because it is closed before called.
        return true;
    };

    template<typename T>
    bool ReplaceBytesEx(const char *name, int id, const T *buffer, unsigned buff_len, unsigned total_len){ // replace or add the record.
        // suppose there are not only a matched records, only 1st could success. in this case, del all then add instead.
        if (!db_name || fmode || !this->Open("r+") || (id==-1)) return false;
        FindStart();
        bool r=FindNext(name, id, 1);
        while (r && FindNext(name, id, 1));
        fh.seek(0, SeekEnd);
        WriteBytes(name, id, buffer, buff_len, total_len);
        this->Close(); // because it is closed before called.
        return true;
    };

    template<typename T>
    void ReplaceBytesContEx(const T *buffer, unsigned length){ // only runs along with ReplaceBytesEx
        this->Open();
        fh.write((const unsigned char*)buffer, sizeof(T)*length);
        this->Close();
    };

    unsigned FindIdNextEx(int id){ // return the name length. and fpos=name pos. FindStart first, then FindIdNextEx(s) until return 0.
        return FindNext("", id, 4);
    };

    void PrintIdEx(int id){
        unsigned len;
        FindStart();
        printf("\r\nid=%d:\r\n", id);
        while (len=FindIdNextEx(id)){
            char r[len+1];
            r[len]=0;
            fh.readBytes(r, len);
            printf("[%s]\r\n", r);
        }
    };

    int FindMaxIdEx(const char *name){ // search the same name of the records in db where has max id.
        // purpose probably is the id for next increments; if there is no RTC timestamps.
        // fpos is the target data.
        // if not found, return -1.
        // if multiple same ones, target is the 1st occurrence.
        int id, maxid;
        unsigned pos;
        FindStart();
        if (!(pos=FindNext(name))) return -1;
        fh.readBytes((char*)&maxid, sizeof(maxid));
        pos=fh.position();
        while (FindNext(name)){
            fh.readBytes((char*)&id, sizeof(id));
            if (maxid<id){
                maxid=id;
                pos=fh.position();
            }
        }
        fh.seek(pos, SeekSet);
        return maxid;
    };

    static void Cp(const char *src, const char *dst){
        if (!filesystem_opened) FILESYSTEM.begin(); // not reasonable but must.
        File fsrc, fdst;
        int f=0;
        if (!(fsrc=FILESYSTEM.open(src, "r"))) f=1;
        if (!f && !(fdst=FILESYSTEM.open(dst, "w"))) f=2;
        if (!f){
            char r[32], n;
            while (n=fsrc.readBytes(r, 32)) fdst.write(r, n);
            fsrc.close();
            fdst.close();
        }
        else if (f==2) fsrc.close();
        if (!filesystem_opened) FILESYSTEM.end(); // not reasonable but must.
    };

    static void Ls(const char *path="/"){
        if (!filesystem_opened) FILESYSTEM.begin(); // not reasonable but must.

        Dir dir = FILESYSTEM.openDir(path);
        printf("list directory \"%s\":\r\n--------\r\n", path);
        while (dir.next())
            printf("%s    %s\r\n", dir.fileName().c_str(), dir.isDirectory()? "<DIR>": (String(dir.fileSize())+" bytes").c_str());

        FSInfo fs_info;
        if (FILESYSTEM.info(fs_info)){
            printf("\r\nentire space %d bytes\r\n%d bytes used\r\nfree space %d bytes\r\n--------\r\n",\
                fs_info.totalBytes, fs_info.usedBytes, fs_info.totalBytes-fs_info.usedBytes);
        }

        if (!filesystem_opened) FILESYSTEM.end(); // not reasonable but must.
    };

} cDataBase;


// ---------------------------------------------- in Esp8266Vars2File.cpp
unsigned cDataBase::filesystem_opened=0;
// ---------------------------------------------- in Esp8266Vars2File.cpp


#endif // _c_VARS_2_FILE_H_
